//
//  AppDelegate.m
//  CA Animation Capture
//
//  Created by Matt Long on 2/2/09.
//  Copyright 2009 Skye Road Systems, Inc.. All rights reserved.
//

#import "AppDelegate.h"

#define SAVEPATH @"/Users/mozilla/Documents/q"
#define QTMOVIEPATH @"/Users/mozilla/Downloads/CA Animation Capture/Sample.mov"
#define USEQTMOVIE false    //This layer crashes on set_visible
#define USENSVIEW false      //Rendering should be possible without using an NSView, but without it we only get the first frame

@implementation AppDelegate

- (CALayer*)createTextLayer;
{
  CATextLayer* txtLayer = [[CATextLayer alloc]init];
  txtLayer.string = [NSString stringWithCString:"CATextLayer!" encoding:[NSString defaultCStringEncoding]];
  return txtLayer;
}
- (CALayer*)createQTLayer;
{
  QTMovie* mov = [[QTMovie movieWithFile:QTMOVIEPATH error:nil] retain];
  [mov play];
  QTMovieLayer* qtLayer = [[QTMovieLayer layerWithMovie:mov] retain];
  return qtLayer;
}
- (void)awakeFromNib;
{
  [[window contentView] setWantsLayer:YES];

  if (!USEQTMOVIE) {
    layer = [self createTextLayer];
  } else {
    layer = [self createQTLayer];
  }
  [layer setDelegate:self];
  [layer setBounds:CGRectMake(0, 0, 10, 10)];
  layer.frame = CGRectMake(0, 0, 400, 400);
  [layer setPosition:CGPointMake(40.0, 40.0)];
  [layer setBorderWidth:1.0];

  CGColorRef bgcolor = CGColorCreateGenericRGB(1.00, 0.0, 0.0, 1.0);
  [layer setBackgroundColor:bgcolor];  
  CFRelease(bgcolor);  
  
  CGColorRef bordercolor = CGColorCreateGenericRGB(0.0, 1.0, 0.0, 1.00);
  [layer setBorderColor:bordercolor];
  CFRelease(bordercolor);
  
  //[[window contentView] setHidden:YES];
  if (USENSVIEW) [[[window contentView] layer] addSublayer:layer];
  
  CGRect frame = [layer frame];
  int imageWidth = ceil(frame.size.width);
  int imageHeight = ceil(frame.size.height);
  int imageSize = imageWidth*imageHeight*4;  
  NSOpenGLPixelFormatAttribute attr[] = {NSOpenGLPFAOffScreen, NSOpenGLPFAColorSize, 32, NSOpenGLPFADepthSize, 24, NSOpenGLPFAAlphaSize, 8, (CGLPixelFormatAttribute) 0};
  NSOpenGLPixelFormat *format = [[NSOpenGLPixelFormat alloc] initWithAttributes:attr];
  g = [[[NSOpenGLContext alloc] initWithFormat:format shareContext:nil] retain];
  caRenderer = [[CARenderer rendererWithCGLContext:[g CGLContextObj] options:NULL] retain];
  
  void* glData = malloc(imageSize);
  [g setOffScreen:glData width:imageWidth height:imageHeight rowbytes:imageWidth*4];
  [g retain];
  
  //caRenderer.layer = layer;

}

- (IBAction)saveAnimation:(id)sender;
{
  NSSavePanel *savePanel;
  
  savePanel = [NSSavePanel savePanel];
  [savePanel setExtensionHidden:YES];
  [savePanel setCanSelectHiddenExtension:NO];
  [savePanel setTreatsFilePackagesAsDirectories:NO];
  
  //if( [savePanel runModal] == NSOKButton )
  //{
    //TODO: Change path
    movie = [[QTMovie alloc] initToWritableFile:SAVEPATH error:nil];    
  //}
  
  [self loadAnimations];

  timer = [NSTimer scheduledTimerWithTimeInterval:1.0/(NSTimeInterval)20.0
                                           target:self
                                         selector:@selector(updateTime:)
                                         userInfo:NULL
                                          repeats:YES];    
  
}

- (void)loadAnimations;
{
  CAAnimationGroup *group = [CAAnimationGroup animation];
  
  [group setAnimations:[NSArray arrayWithObjects:[self backgroundColorAnimation],
                          [self borderWidthAnimation],
                        [self positionAnimation],
                          [self boundsAnimation], nil]];
  
  [group setValue:@"mainGroup" forKey:@"name"];
  [group setDuration:3.0];
  [group setAutoreverses:YES];
  [group setDelegate:self];
  
  [layer addAnimation:group forKey:@"group"];
}

- (void)updateTime:(NSTimer*)theTimer;
{
  NSBitmapImageRep *image = [self getCurrentFrame];
  
  QTTime time = QTMakeTime(1, 20);
  NSDictionary *attrs = [NSDictionary dictionaryWithObject:@"png " forKey:QTAddImageCodecType];
  NSImage *img = [[NSImage alloc] initWithData:[image TIFFRepresentation]];
  [movie addImage:img forDuration:time withAttributes:attrs];
  
  [image release];
}

- (NSBitmapImageRep*)getCurrentFrame;
{
  bool isUseNSView = USENSVIEW;
  //[layer removeFromSuperlayer];
  //[layer retain];
  
  //layer = [layer presentationLayer];
  //[CATransaction commit];
  //[layer setBounds:CGRectMake(0, 0, 10, 10)];
  //layer.frame = CGRectMake(0, 0, 400, 400);
  //[layer setPosition:CGPointMake(40.0, 40.0)];
  NSLog(@"Setting CARenderer");
  @try {
    if (isUseNSView) {
      [layer removeFromSuperlayer];
      [CATransaction commit];
      caRenderer.layer = [layer presentationLayer];
    } else {
      [CATransaction commit];
      CALayer* pres = [layer presentationLayer];
      if (pres != nil) {
        printf("presentationLayer = 0x%X\n", pres);
        caRenderer.layer = pres;
      } else {
        caRenderer.layer = layer;
      }
    }
  }
  
  @catch (NSException *exception) {
    
    NSLog(@"main: Caught %@: %@", [exception name], [exception reason]);
    
  }
  NSLog(@"Set CARenderer");
  
  CGContextRef    context = NULL;
  CGColorSpaceRef colorSpace;
  int bitmapByteCount;
  int bitmapBytesPerRow;
  CGRect frame = [layer frame];
  
  int pixelsHigh = ceil(400);
  int pixelsWide = ceil(400);

  bitmapBytesPerRow   = (pixelsWide * 4);
  bitmapByteCount     = (bitmapBytesPerRow * pixelsHigh);
  
  colorSpace = CGColorSpaceCreateWithName(kCGColorSpaceGenericRGB);
  
  context = CGBitmapContextCreate (NULL,
                                   pixelsWide,
                                   pixelsHigh,
                                   8,
                                   bitmapBytesPerRow,
                                   colorSpace,
                                   kCGImageAlphaPremultipliedLast);
  if (context== NULL)
  {
    NSLog(@"Failed to create context.");
    return nil;
  }
  
  CGColorSpaceRelease( colorSpace );
  
  int imageWidth = ceil(400);
  int imageHeight = ceil(400);
  //[parentLayer setBounds:CGRectMake(0, 0, imageWidth, imageHeight)];
  //[parentLayer setPosition:CGPointMake(0.0, 0.0)];

  //Setup the GLContext
  NSOpenGLContext *oldContext = [NSOpenGLContext currentContext];
  [oldContext retain];
  [g makeCurrentContext];
  glClearColor(0, 0.0, 0.0, 0.0);
  glViewport(0.0, 0.0, imageWidth, imageHeight);
  glLoadIdentity();
  glOrtho (0, imageWidth, 0, imageHeight, -1, 1);
  glMatrixMode(GL_PROJECTION);   glLoadIdentity();
  glMatrixMode(GL_MODELVIEW);    glLoadIdentity();
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  //glScalef(1.f/imageWidth,-1.f/imageHeight,1);
  //glScalef( 2.f/(imageWidth),-2.f/(imageHeight), 1.0f );
  [CATransaction commit];
  
  //Draw triangle to test glContext
  glColor3f(rand() % 1000 / 1000.0f,0.0f,0.0f);
  glBegin(GL_TRIANGLES);
    glVertex3f( 0.0f, 1.0f, 0.0f);
    glColor3f( 0.0f,1.0f,0.0f);
    glVertex3f(-1.0f,-1.0f, 0.0f);
    glColor3f( 1.0f,1.0f,1.0f);
    glVertex3f( 1.0f,-1.0f, 0.0f);
  glEnd();
  glTranslatef(-0.5f,1.f, 0.0f);
  glScalef( 1.f/(imageWidth),-1.f/(imageHeight), 1.0f );

  
  //Draw the current frame.
  //caRenderer.layer = [[ [ window contentView] layer] presentationLayer];
  [CATransaction commit];
  [layer setNeedsDisplay];
  caRenderer.bounds = CGRectMake(-imageWidth*2000,-imageHeight*2000,imageWidth*200000,imageHeight*200000);
  static double c = 0;
  c+=1;
  if( c > 20 ) {
    [movie updateMovieFile];
    [timer invalidate];
  }
  //CGLLockContext([g CGLContextObh]);
  [caRenderer beginFrameAtTime:(c) timeStamp:NULL];
  [caRenderer addUpdateRect:CGRectMake(-1000,-1000,10000,10000)];
  [caRenderer render];
  [caRenderer endFrame];
  //CGLUnlockContext([g CGLContextObj]);
  glFlush();
  [g flushBuffer];
  
  glPixelStorei(GL_PACK_ALIGNMENT, 4);
  glPixelStorei(GL_PACK_ROW_LENGTH, 0);
  glPixelStorei(GL_PACK_SKIP_ROWS, 0);
  glPixelStorei(GL_PACK_SKIP_PIXELS, 0);
  
  //CGImageRef img = CGBitmapContextCreateImage(context);
  //NSBitmapImageRep *bitmap = [[NSBitmapImageRep alloc] initWithCGImage:img];
  //CFRelease(img);
  
  NSBitmapImageRep *bitmap = [[NSBitmapImageRep alloc]
            initWithBitmapDataPlanes:(unsigned char**)NULL
            pixelsWide:frame.size.width
            pixelsHigh:frame.size.height
            bitsPerSample:8
            samplesPerPixel:4
            hasAlpha:YES
            isPlanar:NO
            colorSpaceName:NSCalibratedRGBColorSpace
            bytesPerRow:4 * frame.size.width
            bitsPerPixel:32];
  
  
  glReadPixels(0.0f, 0.0f, frame.size.width,frame.size.height,
               GL_RGBA, GL_UNSIGNED_BYTE,
               [bitmap bitmapData]);
  
  
  //[[[[window contentView] layer] presentationLayer] renderInContext:context];
  //CGImageRef img = CGBitmapContextCreateImage(context);
  //bitmap = [[NSBitmapImageRep alloc] initWithCGImage:img];
  //CFRelease(img);
  
  if (isUseNSView) {
    caRenderer.layer = NULL;
    [CATransaction commit];
    [[[window contentView] layer] addSublayer:layer];
    [CATransaction commit];
  } else {
    
    caRenderer.layer = NULL;
    [CATransaction commit];
  }
  return bitmap;

}

- (void)animationDidStop:(CAAnimation *)theAnimation finished:(BOOL)flag;
{
  NSLog(@"Animation stopped: %@", theAnimation);
  
  id name = [theAnimation valueForKey:@"name"];
  if( name )
    if( [name isEqualToString:@"mainGroup"] )
    {
      [movie updateMovieFile];
      //[timer invalidate];
    }
}


#pragma mark -
#pragma mark Animations

- (CAAnimation*)positionAnimation;
{
  CGMutablePathRef thePath = CGPathCreateMutable();
  CGPathMoveToPoint(thePath,NULL,74.0,74.0);
  CGPathAddCurveToPoint(thePath,NULL,74.0,500.0,
                        320.0,500.0,
                        320.0,74.0);

  CGPathAddCurveToPoint(thePath,NULL,320.0,500.0,
                        566.0,500.0,
                        566.0,74.0);
  
  CAKeyframeAnimation *animation = [CAKeyframeAnimation animationWithKeyPath:@"position"];
  
  [animation setPath:thePath];
  [animation setDuration:5.0];
  [animation setAutoreverses:YES];
  [animation setRotationMode:kCAAnimationRotateAuto];
  [animation setFillMode:kCAFillModeForwards];
  
  CFRelease(thePath);
  
  return animation;
}

- (CAAnimation*)backgroundColorAnimation;
{
	CAKeyframeAnimation *animation = [CAKeyframeAnimation animationWithKeyPath:@"backgroundColor"];
  
	[animation setDuration:5.0];

	NSMutableArray *values = [NSMutableArray array];
  
  CGColorRef red = CGColorCreateGenericRGB(1.0, 0.0, 0.0, 1.0);
  CGColorRef green = CGColorCreateGenericRGB(0.0, 1.0, 0.0, 1.0);
  CGColorRef blue = CGColorCreateGenericRGB(0.0, 0.0, 1.0, 1.0);
  
  [values addObject:(NSColor*)red];
  [values addObject:(NSColor*)green];
  [values addObject:(NSColor*)blue];

	[animation setValues:values];
  [animation setCalculationMode:kCAAnimationLinear];
  [animation setAutoreverses:YES];

	return animation;
}

- (CAAnimation*)borderWidthAnimation;
{
	CAKeyframeAnimation *animation = [CAKeyframeAnimation animationWithKeyPath:@"borderWidth"];
  
	[animation setDuration:5.0];
  
	NSMutableArray *values = [NSMutableArray array];
  
  [values addObject:[NSNumber numberWithFloat:3.25]];
  [values addObject:[NSNumber numberWithFloat:11.25]];
  [values addObject:[NSNumber numberWithFloat:40.25]];
  [values addObject:[NSNumber numberWithFloat:5.25]];
  [values addObject:[NSNumber numberWithFloat:18.0]];
  
	[animation setValues:values];
  [animation setAutoreverses:YES];

	return animation;
}

- (CAAnimation*)boundsAnimation;
{
	CAKeyframeAnimation *animation = [CAKeyframeAnimation animationWithKeyPath:@"bounds"];
  
	[animation setDuration:5.0]; 
  
	NSMutableArray *values = [NSMutableArray array];
  
  NSRect startRect = NSMakeRect(10.0, 10.0, 50.0, 50.0);
  NSRect nextRect = NSMakeRect(300.0, 300.0, 200.0, 200.0);
  NSRect lastRect = NSMakeRect(200.0, 200.0, 300.0, 300.0);
  NSRect r = NSMakeRect(10.0, 10.0, 50.0, 50.0);
    
  [values addObject:[NSValue valueWithRect:startRect]];
  [values addObject:[NSValue valueWithRect:nextRect]];
  [values addObject:[NSValue valueWithRect:lastRect]];
  [values addObject:[NSValue valueWithRect:r]];
  
	[animation setValues:values];
  [animation setAutoreverses:YES];
  
	return animation;
}


- (void)dealloc;
{
  [movie release], movie = nil;
  [super dealloc];
}

@end
